import asyncio
from collections.abc import AsyncIterator, Callable
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import create_autospec

import pytest
from contentctl.plan import (
    SyncAction,
    SyncError,
    SyncFilters,
    SyncPolicy,
    SyncScope,
    plan_sync,
)
from contentctl.plan.sync import SyncOperation

if TYPE_CHECKING:
    from tests.fixture_types import FixturePath

type PlanPolicy = Callable[[], SyncPolicy]
type CollectOperations = Callable[[AsyncIterator[SyncOperation]], list[SyncOperation]]


@pytest.mark.parametrize(
    ("source", "dest", "path"),
    [
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            str(Path("/") / "abs"),
        ),
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            "../escape",
        ),
        (("nonexistent_source",), ("nonexistent_destination",), "."),
        (("plan_sync", "source_multi"), ("plan_sync", "source_multi"), "."),
        (("plan_sync", "source_multi"), ("plan_sync", "source_multi", "sub"), "."),
        (("plan_sync", "source_multi", "sub"), ("plan_sync", "source_multi"), "."),
    ],
)
def test_plan_sync_rejects_invalid_inputs(
    fixture_path: FixturePath,
    source: tuple[str, ...],
    dest: tuple[str, ...],
    path: str,
) -> None:
    with pytest.raises(SyncError):
        plan_sync(
            SyncScope(
                source_root=fixture_path(*source),
                destination_root=fixture_path(*dest),
                path=path,
            ),
            SyncFilters(
                source_include=(),
                source_exclude=(),
                destination_include=(),
                destination_exclude=(),
            ),
            policy=SyncPolicy(semaphore=asyncio.Semaphore(1)),
        )


@pytest.mark.parametrize(
    (
        "source",
        "dest",
        "source_include",
        "source_exclude",
        "destination_include",
        "destination_exclude",
        "sync_path",
        "expected_files",
    ),
    [
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            ("*.txt", "**/*.txt"),
            ("sub/*",),
            ("*.txt", "**/*.txt"),
            (),
            ".",
            {"guide.txt"},
        ),
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            ("sub/**",),
            (),
            (),
            (),
            "sub",
            {"chapter.txt"},
        ),
    ],
)
def test_plan_sync_applies_include_exclude(
    fixture_path: FixturePath,
    plan_policy: PlanPolicy,
    collect_operations: CollectOperations,
    source: tuple[str, ...],
    dest: tuple[str, ...],
    source_include: tuple[str, ...],
    source_exclude: tuple[str, ...],
    destination_include: tuple[str, ...],
    destination_exclude: tuple[str, ...],
    sync_path: str,
    expected_files: set[str],
) -> None:
    source_root = fixture_path(*source)
    destination_root = fixture_path(*dest)

    plan = plan_sync(
        SyncScope(
            source_root=source_root, destination_root=destination_root, path=sync_path
        ),
        SyncFilters(
            source_include=source_include,
            source_exclude=source_exclude,
            destination_include=destination_include,
            destination_exclude=destination_exclude,
        ),
        policy=plan_policy(),
    )

    collected = collect_operations(plan.stream)
    paths = {op.relative.name for op in collected}

    assert paths == expected_files


@pytest.mark.parametrize(
    (
        "source",
        "dest",
        "filename",
        "source_exclude",
        "expected_count",
        "expected_action",
    ),
    [
        (
            ("plan_sync", "source_single"),
            ("plan_sync", "destination_single"),
            "note.txt",
            (),
            1,
            SyncAction.REPLACE,
        ),
        (
            ("plan_sync", "source_single"),
            ("plan_sync", "destination_single"),
            "note.txt",
            ("*.txt",),
            0,
            None,
        ),
    ],
)
def test_plan_sync_file_source(
    fixture_path: FixturePath,
    plan_policy: PlanPolicy,
    collect_operations: CollectOperations,
    make_sync_op: Callable[[str, SyncAction], SyncOperation],
    source: tuple[str, ...],
    dest: tuple[str, ...],
    filename: str,
    source_exclude: tuple[str, ...],
    expected_count: int,
    expected_action: SyncAction | None,
) -> None:
    source_root = fixture_path(*source)
    destination_root = fixture_path(*dest)

    plan = plan_sync(
        SyncScope(
            source_root=source_root, destination_root=destination_root, path=filename
        ),
        SyncFilters(
            source_include=(),
            source_exclude=source_exclude,
            destination_include=(),
            destination_exclude=(),
        ),
        policy=plan_policy(),
    )

    collected = collect_operations(plan.stream)
    if expected_count > 0:
        assert expected_action is not None
        expected = [make_sync_op(filename, expected_action)]
        assert collected == expected
    else:
        assert collected == []


@pytest.mark.parametrize(
    (
        "source",
        "dest",
        "destination_include",
        "destination_exclude",
        "sync_path",
        "expected_actions",
    ),
    [
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            ("*.txt", "**/*.txt"),
            ("sub/*",),
            ".",
            {
                "guide.txt": SyncAction.COPY,
                "readme.md": SyncAction.SKIP,
                "sub/chapter.txt": SyncAction.SKIP,
            },
        ),
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            (),
            ("*.txt", "**/*.txt"),
            ".",
            {
                "guide.txt": SyncAction.SKIP,
                "sub/chapter.txt": SyncAction.SKIP,
                "readme.md": SyncAction.COPY,
            },
        ),
    ],
)
def test_plan_sync_destination_filtering(
    fixture_path: FixturePath,
    plan_policy: PlanPolicy,
    collect_operations: CollectOperations,
    source: tuple[str, ...],
    dest: tuple[str, ...],
    destination_include: tuple[str, ...],
    destination_exclude: tuple[str, ...],
    sync_path: str,
    expected_actions: dict[str, SyncAction],
) -> None:
    source_root = fixture_path(*source)
    destination_root = fixture_path(*dest)
    plan = plan_sync(
        SyncScope(
            source_root=source_root, destination_root=destination_root, path=sync_path
        ),
        SyncFilters(
            source_include=(),
            source_exclude=(),
            destination_include=destination_include,
            destination_exclude=destination_exclude,
        ),
        policy=plan_policy(),
    )

    actions = {str(op.relative): op.action for op in collect_operations(plan.stream)}

    expected = dict(expected_actions)
    assert {k: v for k, v in actions.items() if k in expected} == expected


@pytest.mark.parametrize(
    ("source", "dest", "sync_path", "expected_actions"),
    [
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_with_extra"),
            ".",
            {
                "old.txt": SyncAction.DELETE,
                "guide.txt": SyncAction.REPLACE,
            },
        ),
    ],
)
def test_plan_sync_with_delete_removes_unmanaged_files(
    fixture_path: FixturePath,
    plan_policy: PlanPolicy,
    collect_operations: CollectOperations,
    source: tuple[str, ...],
    dest: tuple[str, ...],
    sync_path: str,
    expected_actions: dict[str, SyncAction],
) -> None:
    source_root = fixture_path(*source)
    destination_root = fixture_path(*dest)

    policy = plan_policy()
    plan = plan_sync(
        SyncScope(
            source_root=source_root, destination_root=destination_root, path=sync_path
        ),
        SyncFilters(
            source_include=(),
            source_exclude=(),
            destination_include=(),
            destination_exclude=(),
        ),
        policy=SyncPolicy(semaphore=policy.semaphore, allow_delete=True),
    )

    collected = collect_operations(plan.stream)
    actions = {str(op.relative): op.action for op in collected}

    assert all(actions.get(file) is action for file, action in expected_actions.items())


@pytest.mark.parametrize(
    ("source", "dest", "sync_path", "missing_file", "expected_file", "expected_action"),
    [
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_with_extra"),
            ".",
            "old.txt",
            "guide.txt",
            SyncAction.REPLACE,
        ),
    ],
)
def test_plan_sync_without_delete_keeps_extra_files(
    fixture_path: FixturePath,
    plan_policy: PlanPolicy,
    collect_operations: CollectOperations,
    source: tuple[str, ...],
    dest: tuple[str, ...],
    sync_path: str,
    missing_file: str,
    expected_file: str,
    expected_action: SyncAction,
) -> None:
    source_root = fixture_path(*source)
    destination_root = fixture_path(*dest)

    plan = plan_sync(
        SyncScope(
            source_root=source_root, destination_root=destination_root, path=sync_path
        ),
        SyncFilters(
            source_include=(),
            source_exclude=(),
            destination_include=(),
            destination_exclude=(),
        ),
        policy=plan_policy(),
    )

    collected = collect_operations(plan.stream)
    actions = {str(op.relative): op.action for op in collected}

    assert missing_file not in actions
    assert actions[expected_file] is expected_action


@pytest.mark.parametrize(
    (
        "source",
        "dest",
        "source_include",
        "destination_include",
        "sync_path",
        "expected_actions",
    ),
    [
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_with_extra"),
            ("*.txt", "**/*.txt"),
            ("*.txt", "**/*.txt"),
            ".",
            {
                "old.txt": SyncAction.DELETE,
                "guide.txt": SyncAction.REPLACE,
                "readme.md": None,
            },
        ),
    ],
)
def test_plan_sync_with_delete_respects_selector_intersection(
    fixture_path: FixturePath,
    plan_policy: PlanPolicy,
    collect_operations: CollectOperations,
    source: tuple[str, ...],
    dest: tuple[str, ...],
    source_include: tuple[str, ...],
    destination_include: tuple[str, ...],
    sync_path: str,
    expected_actions: dict[str, SyncAction | None],
) -> None:
    source_root = fixture_path(*source)
    destination_root = fixture_path(*dest)

    policy = plan_policy()
    plan = plan_sync(
        SyncScope(
            source_root=source_root, destination_root=destination_root, path=sync_path
        ),
        SyncFilters(
            source_include=source_include,
            source_exclude=(),
            destination_include=destination_include,
            destination_exclude=(),
        ),
        policy=SyncPolicy(semaphore=policy.semaphore, allow_delete=True),
    )

    collected = collect_operations(plan.stream)
    actions = {str(op.relative): op.action for op in collected}

    assert all(
        (file not in actions) if action is None else (actions.get(file) is action)
        for file, action in expected_actions.items()
    )


@pytest.mark.parametrize(
    ("source", "dest", "filename"),
    [
        (
            ("plan_sync", "source_single"),
            ("plan_sync", "destination_single"),
            "note.txt",
        ),
    ],
)
def test_plan_sync_file_source_copies_when_destination_missing(
    fixture_path: FixturePath,
    plan_policy: PlanPolicy,
    collect_operations: CollectOperations,
    make_sync_op: Callable[[str, SyncAction], SyncOperation],
    monkeypatch: pytest.MonkeyPatch,
    source: tuple[str, ...],
    dest: tuple[str, ...],
    filename: str,
) -> None:
    source_root = fixture_path(*source)
    destination_root = fixture_path(*dest)

    destination_path = destination_root / filename

    def exists_stub(self: Path) -> bool:
        return self != destination_path

    exists_mock = create_autospec(Path.exists, side_effect=exists_stub)
    monkeypatch.setattr(Path, "exists", exists_mock)

    plan = plan_sync(
        SyncScope(
            source_root=source_root, destination_root=destination_root, path=filename
        ),
        SyncFilters(
            source_include=(),
            source_exclude=(),
            destination_include=(),
            destination_exclude=(),
        ),
        policy=plan_policy(),
    )

    collected = collect_operations(plan.stream)
    assert collected == [make_sync_op(filename, SyncAction.COPY)]


@pytest.fixture
def plan_policy() -> PlanPolicy:
    """Create semaphores for async plan operations."""

    def _make() -> SyncPolicy:
        return SyncPolicy(semaphore=asyncio.Semaphore(2))

    return _make


@pytest.fixture
def collect_operations() -> CollectOperations:
    """Collect operations from async iterator."""

    def _collect(operations: AsyncIterator[SyncOperation]) -> list[SyncOperation]:
        async def collect() -> list[SyncOperation]:
            return [operation async for operation in operations]

        return asyncio.run(collect())

    return _collect



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, TextIO

from contentctl.operation_kit import execute_sync_operation
from contentctl.plan.sync import SyncFilters, SyncPolicy, SyncScope, plan_sync
from contentctl.utils.concurrent import default_concurrency

if TYPE_CHECKING:
    from collections.abc import Iterable

    from contentctl.config import Workspace


async def run_deploy(
    workspaces: Iterable[Workspace],
    origin: Workspace,
    path: str,
    output: TextIO,
    *,
    dry_run: bool,
    verbose: bool,
    allow_delete: bool,
) -> None:
    base = default_concurrency()
    io_semaphore = asyncio.Semaphore(base)

    for workspace in workspaces:
        plan = plan_sync(
            SyncScope(origin.path, workspace.path, path),
            SyncFilters(
                source_include=origin.include,
                source_exclude=origin.exclude,
                destination_include=workspace.include,
                destination_exclude=workspace.exclude,
            ),
            policy=SyncPolicy(semaphore=io_semaphore, allow_delete=allow_delete),
        )

        if verbose or dry_run:
            print(
                f"deploy {workspace.name}: {plan.source_path} -> {plan.destination_path}",
                file=output,
            )

        await execute_sync_operation(
            stream=plan.stream,
            source_root=plan.source_root,
            destination_root=plan.destination_root,
            semaphore=io_semaphore,
            operation_name="deploy",
            workspace_name=workspace.name,
            dry_run=dry_run,
            verbose=verbose,
            output=output,
        )



import asyncio
import shutil
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import create_autospec

import pytest
from contentctl.execute.sync import apply_sync_plan, print_sync_plan
from contentctl.plan.sync import SyncAction, SyncOperation

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Callable


@pytest.mark.parametrize(
    ("operations", "expected_copies", "expected_deletions"),
    [
        (
            [("guide.txt", SyncAction.COPY), ("drafts.txt", SyncAction.SKIP)],
            [("guide.txt", "guide.txt")],
            [],
        ),
        (
            [("readme.md", SyncAction.REPLACE), ("index.html", SyncAction.SKIP)],
            [("readme.md", "readme.md")],
            [],
        ),
        (
            [("old.txt", SyncAction.DELETE), ("keep.txt", SyncAction.SKIP)],
            [],
            ["old.txt"],
        ),
        (
            [("a.txt", SyncAction.DELETE), ("b.txt", SyncAction.DELETE)],
            [],
            ["a.txt", "b.txt"],
        ),
    ],
)
def test_apply_sync_plan_applies_actions(
    monkeypatch: pytest.MonkeyPatch,
    source_root: Path,
    destination_root: Path,
    make_sync_op: Callable[[str, SyncAction], SyncOperation],
    make_stream: Callable[..., AsyncIterator[SyncOperation]],
    drain_stream: Callable[[AsyncIterator[SyncOperation]], None],
    operations: list[tuple[str, SyncAction]],
    expected_copies: list[tuple[str, str]],
    expected_deletions: list[str],
) -> None:
    observed_copies: list[tuple[Path, Path]] = []
    observed_deletions: list[Path] = []

    def observe_copy(src: Path, dst: Path) -> None:
        observed_copies.append((src, dst))

    def observe_unlink(self: Path) -> None:
        observed_deletions.append(self)

    def exists_stub(_self: Path) -> bool:
        return True

    copy2_mock = create_autospec(shutil.copy2, side_effect=observe_copy)
    mkdir_mock = create_autospec(Path.mkdir)
    unlink_mock = create_autospec(Path.unlink, side_effect=observe_unlink)
    exists_mock = create_autospec(Path.exists, side_effect=exists_stub)

    monkeypatch.setattr("contentctl.execute.sync.shutil.copy2", copy2_mock)
    monkeypatch.setattr(Path, "mkdir", mkdir_mock)
    monkeypatch.setattr(Path, "unlink", unlink_mock)
    monkeypatch.setattr(Path, "exists", exists_mock)

    stream = make_stream(*[make_sync_op(path, action) for path, action in operations])

    observed = apply_sync_plan(
        stream,
        source_root=source_root,
        destination_root=destination_root,
        semaphore=asyncio.Semaphore(2),
    )
    drain_stream(observed)

    expected_copy_paths = [
        (source_root / src, destination_root / dst) for src, dst in expected_copies
    ]
    expected_delete_paths = [destination_root / path for path in expected_deletions]
    assert observed_copies == expected_copy_paths
    assert observed_deletions == expected_delete_paths


@pytest.mark.parametrize(
    "filename",
    [
        "ghost.txt",
        "missing.md",
    ],
)
def test_apply_sync_plan_skips_delete_when_missing(
    monkeypatch: pytest.MonkeyPatch,
    source_root: Path,
    destination_root: Path,
    make_sync_op: Callable[[str, SyncAction], SyncOperation],
    make_stream: Callable[..., AsyncIterator[SyncOperation]],
    drain_stream: Callable[[AsyncIterator[SyncOperation]], None],
    filename: str,
) -> None:
    def exists_stub(_self: Path) -> bool:
        return False

    copy2_mock = create_autospec(shutil.copy2)
    mkdir_mock = create_autospec(Path.mkdir)
    unlink_mock = create_autospec(Path.unlink)
    exists_mock = create_autospec(Path.exists, side_effect=exists_stub)

    monkeypatch.setattr("contentctl.execute.sync.shutil.copy2", copy2_mock)
    monkeypatch.setattr(Path, "mkdir", mkdir_mock)
    monkeypatch.setattr(Path, "unlink", unlink_mock)
    monkeypatch.setattr(Path, "exists", exists_mock)

    stream = make_stream(make_sync_op(filename, SyncAction.DELETE))

    observed = apply_sync_plan(
        stream,
        source_root=source_root,
        destination_root=destination_root,
        semaphore=asyncio.Semaphore(2),
    )
    drain_stream(observed)

    unlink_mock.assert_not_called()
    copy2_mock.assert_not_called()
    mkdir_mock.assert_not_called()


@pytest.mark.parametrize(
    ("operations", "expected_lines"),
    [
        (
            [("guide.txt", SyncAction.COPY), ("drafts.txt", SyncAction.SKIP)],
            [
                "copy    /virtual/destination/guide.txt",
                "skip    /virtual/destination/drafts.txt",
            ],
        ),
        (
            [("readme.md", SyncAction.REPLACE)],
            [
                "replace /virtual/destination/readme.md",
            ],
        ),
        (
            [("old.txt", SyncAction.DELETE)],
            [
                "delete  /virtual/destination/old.txt",
            ],
        ),
    ],
)
def test_print_sync_plan_formats_lines(
    destination_root: Path,
    make_sync_op: Callable[[str, SyncAction], SyncOperation],
    make_stream: Callable[..., AsyncIterator[SyncOperation]],
    drain_stream: Callable[[AsyncIterator[SyncOperation]], None],
    operations: list[tuple[str, SyncAction]],
    expected_lines: list[str],
) -> None:
    stream = make_stream(*[make_sync_op(path, action) for path, action in operations])

    output = StringIO()
    observed = print_sync_plan(
        stream,
        destination_root=destination_root,
        output=output,
    )
    drain_stream(observed)

    lines = output.getvalue().splitlines()
    assert lines == expected_lines


@pytest.fixture
def source_root() -> Path:
    """Virtual source root for sync tests."""
    return Path("/virtual/source")


@pytest.fixture
def destination_root() -> Path:
    """Virtual destination root for sync tests."""
    return Path("/virtual/destination")


@pytest.fixture
def make_stream() -> Callable[..., AsyncIterator[SyncOperation]]:
    """Create an async stream of SyncOperations."""

    def _make(*operations: SyncOperation) -> AsyncIterator[SyncOperation]:
        async def iter_operations() -> AsyncIterator[SyncOperation]:
            for operation in operations:
                yield operation

        return iter_operations()

    return _make


@pytest.fixture
def drain_stream() -> Callable[[AsyncIterator[SyncOperation]], None]:
    """Drain an async stream to completion."""

    def _drain(stream: AsyncIterator[SyncOperation]) -> None:
        async def consume() -> None:
            async for _ in stream:
                pass

        asyncio.run(consume())

    return _drain



"""Type definitions for shared pytest fixtures.

This module provides type definitions for fixtures that are used
across multiple test modules.
"""

from collections.abc import Callable
from pathlib import Path

type FixturePath = Callable[..., Path]



from .sync import (
    SyncAction,
    SyncError,
    SyncFilters,
    SyncOperation,
    SyncPlan,
    SyncPolicy,
    SyncScope,
    plan_sync,
)

__all__ = [
    "SyncAction",
    "SyncError",
    "SyncFilters",
    "SyncOperation",
    "SyncPlan",
    "SyncPolicy",
    "SyncScope",
    "plan_sync",
]



import sys
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from collections.abc import Callable


@pytest.fixture
def fixture_path() -> Callable[..., Path]:
    """Return a function to get fixture file paths."""

    def _fixture_path(*parts: str) -> Path:
        return FIXTURE_ROOT.joinpath(*parts)

    return _fixture_path


@pytest.fixture
def fixture_dir() -> Path:
    """Return the path to the _fixture directory."""
    return FIXTURE_ROOT


TESTS_ROOT = Path(__file__).resolve().parent
PACKAGE_ROOT = TESTS_ROOT.parent
SRC_ROOT = PACKAGE_ROOT / "src"
FIXTURE_ROOT = TESTS_ROOT / "_fixture"


def _prepend_sys_path(path: Path) -> None:
    """Ensure test and source roots are importable in pytest importlib mode."""
    path_str = str(path)
    if path_str not in sys.path:
        sys.path.insert(0, path_str)


_prepend_sys_path(PACKAGE_ROOT)
_prepend_sys_path(SRC_ROOT)
_prepend_sys_path(TESTS_ROOT)



"""Type definitions for contentctl module fixtures."""

from collections.abc import Callable
from pathlib import Path
from typing import Any

type MakeWorkspace = Callable[[str, str | Path], Any]
type MakeSyncOp = Callable[[str, Any], Any]



from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Protocol
from unittest.mock import create_autospec

import contentctl.__main__ as mainmod
import pytest
from contentctl.config import ResolvedConfig
from contentctl.gateway import AdoptContext, DeployContext, InitContext
from contentctl.gateway.defaults import DEFAULT_CONFIG_FILENAME

if TYPE_CHECKING:
    from .fixture_types import MakeWorkspace

type DeployCtxFactory = Callable[..., DeployContext]
type AdoptCtxFactory = Callable[..., AdoptContext]


class PatchMainContext(Protocol):
    def __call__(
        self, *, ctx: AdoptContext | DeployContext, resolved: ResolvedConfig
    ) -> None: ...


@pytest.fixture
def config_path() -> Path:
    """Config path for main tests."""
    return Path("/config.yaml")


@pytest.fixture
def deploy_ctx(config_path: Path) -> DeployCtxFactory:
    """Create DeployContext for testing."""

    def _make(
        workspaces: list[str] | None = None,
        path: str = ".",
        *,
        all_workspaces: bool = True,
        dry_run: bool = False,
        verbose: bool = False,
        allow_delete: bool = False,
    ) -> DeployContext:
        return DeployContext(
            command="deploy",
            config_path=config_path,
            all_workspaces=all_workspaces,
            workspaces=workspaces or [],
            path=path,
            dry_run=dry_run,
            verbose=verbose,
            allow_delete=allow_delete,
        )

    return _make


@pytest.fixture
def adopt_ctx(config_path: Path) -> AdoptCtxFactory:
    """Create AdoptContext for testing."""

    def _make(
        workspace: str = "alpha",
        path: str = "docs",
        *,
        dry_run: bool = False,
        verbose: bool = False,
    ) -> AdoptContext:
        return AdoptContext(
            command="adopt",
            config_path=config_path,
            workspace=workspace,
            path=path,
            dry_run=dry_run,
            verbose=verbose,
        )

    return _make


@pytest.fixture
def resolved_config(make_workspace: MakeWorkspace) -> ResolvedConfig:
    """Create a resolved config for testing."""
    return ResolvedConfig(
        origin=make_workspace("", "/origin"),
        workspaces={
            "zeta": make_workspace("zeta", "/zeta"),
            "alpha": make_workspace("alpha", "/alpha"),
        },
    )


@pytest.fixture
def patch_main_context(
    monkeypatch: pytest.MonkeyPatch,
) -> PatchMainContext:
    """Patch main context loading functions."""

    def _patch(
        ctx: AdoptContext | DeployContext,
        resolved: ResolvedConfig,
    ) -> None:
        monkeypatch.setattr(
            mainmod,
            "parse_cli",
            create_autospec(mainmod.parse_cli, return_value=ctx),
        )
        monkeypatch.setattr(
            mainmod,
            "load_config",
            create_autospec(mainmod.load_config, return_value={}),
        )
        monkeypatch.setattr(
            mainmod,
            "resolve_config",
            create_autospec(mainmod.resolve_config, return_value=resolved),
        )

    return _patch


@pytest.mark.parametrize(
    ("failure", "message"),
    [
        ("load_config", "bad config"),
        ("resolve_config", "resolve failed"),
    ],
)
def test_main_exits_on_config_errors(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    deploy_ctx: DeployCtxFactory,
    resolved_config: ResolvedConfig,
    failure: str,
    message: str,
) -> None:
    ctx = deploy_ctx()

    def load_config_side_effect(_path: Path) -> dict[str, object]:
        if failure == "load_config":
            raise mainmod.ConfigError(message)
        return {}

    def resolve_config_side_effect(
        _cfg: dict[str, object], _path: Path
    ) -> ResolvedConfig:
        if failure == "resolve_config":
            raise mainmod.ConfigError(message)
        return resolved_config

    parse_cli_mock = create_autospec(mainmod.parse_cli, return_value=ctx)
    load_config_mock = create_autospec(
        mainmod.load_config, side_effect=load_config_side_effect
    )
    resolve_config_mock = create_autospec(
        mainmod.resolve_config, side_effect=resolve_config_side_effect
    )

    monkeypatch.setattr(mainmod, "parse_cli", parse_cli_mock)
    monkeypatch.setattr(mainmod, "load_config", load_config_mock)
    monkeypatch.setattr(mainmod, "resolve_config", resolve_config_mock)

    with pytest.raises(SystemExit) as excinfo:
        mainmod.main()

    assert excinfo.value.code == 2
    assert message in capsys.readouterr().err


@pytest.mark.parametrize(
    "message",
    [
        "sync failed",
        "operation error",
    ],
)
def test_main_exits_on_sync_error(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    deploy_ctx: DeployCtxFactory,
    resolved_config: ResolvedConfig,
    patch_main_context: PatchMainContext,
    message: str,
) -> None:
    ctx = deploy_ctx()
    patch_main_context(ctx=ctx, resolved=resolved_config)

    async def raise_sync_error(*_args: object, **_kwargs: object) -> None:
        raise mainmod.SyncError(message)

    monkeypatch.setattr(mainmod, "run_deploy", raise_sync_error)

    with pytest.raises(SystemExit) as excinfo:
        mainmod.main()

    assert excinfo.value.code == 1
    assert message in capsys.readouterr().err


@pytest.mark.parametrize(
    ("workspaces", "message"),
    [
        (["missing"], "unknown workspace"),
        (["invalid", "notfound"], "unknown workspace"),
    ],
)
def test_main_exits_on_dispatch_config_error(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    deploy_ctx: DeployCtxFactory,
    resolved_config: ResolvedConfig,
    patch_main_context: PatchMainContext,
    workspaces: list[str],
    message: str,
) -> None:
    ctx = deploy_ctx(all_workspaces=False, workspaces=workspaces)

    select_workspaces_mock = create_autospec(
        mainmod.select_workspaces, side_effect=mainmod.ConfigError(message)
    )

    patch_main_context(ctx=ctx, resolved=resolved_config)
    monkeypatch.setattr(mainmod, "select_workspaces", select_workspaces_mock)

    with pytest.raises(SystemExit) as excinfo:
        mainmod.main()

    assert excinfo.value.code == 1
    assert message in capsys.readouterr().err


@pytest.mark.parametrize(
    ("verbose", "delete"),
    [
        (True, True),
        (True, False),
        (False, True),
        (False, False),
    ],
)
def test_main_dispatches_deploy_with_flags(
    monkeypatch: pytest.MonkeyPatch,
    deploy_ctx: DeployCtxFactory,
    resolved_config: ResolvedConfig,
    patch_main_context: PatchMainContext,
    *,
    verbose: bool,
    delete: bool,
) -> None:
    ctx = deploy_ctx(verbose=verbose, allow_delete=delete)

    run_deploy_mock = create_autospec(mainmod.run_deploy)
    monkeypatch.setattr(mainmod, "run_deploy", run_deploy_mock)
    patch_main_context(ctx=ctx, resolved=resolved_config)

    mainmod.main()

    run_deploy_mock.assert_called_once()
    call_kwargs = run_deploy_mock.call_args.kwargs
    assert "workspaces" in call_kwargs
    assert call_kwargs["verbose"] is verbose
    assert call_kwargs["allow_delete"] is delete


@pytest.mark.parametrize(
    ("workspaces", "workspace_root_prefix"),
    [
        (["alpha", "zeta"], "/"),
    ],
)
def test_main_dispatches_deploy_selected_workspaces(
    monkeypatch: pytest.MonkeyPatch,
    deploy_ctx: DeployCtxFactory,
    make_workspace: MakeWorkspace,
    resolved_config: ResolvedConfig,
    patch_main_context: PatchMainContext,
    workspaces: list[str],
    workspace_root_prefix: str,
) -> None:
    ctx = deploy_ctx(all_workspaces=False, workspaces=workspaces)
    selected = [
        make_workspace(name, f"{workspace_root_prefix}{name}") for name in workspaces
    ]

    select_workspaces_mock = create_autospec(
        mainmod.select_workspaces, return_value=selected
    )
    select_all_workspaces_mock = create_autospec(
        mainmod.select_all_workspaces,
        side_effect=AssertionError("select_all_workspaces should not run"),
    )
    run_deploy_mock = create_autospec(mainmod.run_deploy)
    monkeypatch.setattr(mainmod, "select_workspaces", select_workspaces_mock)
    monkeypatch.setattr(mainmod, "select_all_workspaces", select_all_workspaces_mock)
    monkeypatch.setattr(mainmod, "run_deploy", run_deploy_mock)
    patch_main_context(ctx=ctx, resolved=resolved_config)

    mainmod.main()

    select_workspaces_mock.assert_called_once_with(resolved_config, workspaces)
    run_deploy_mock.assert_called_once()


@pytest.mark.parametrize(
    ("workspace_name", "workspace_path", "dry_run", "verbose"),
    [
        ("alpha", "docs", True, True),
        ("alpha", "docs", True, False),
        ("alpha", "docs", False, True),
        ("alpha", "docs", False, False),
    ],
)
def test_main_dispatches_adopt_workspace(
    monkeypatch: pytest.MonkeyPatch,
    adopt_ctx: AdoptCtxFactory,
    resolved_config: ResolvedConfig,
    patch_main_context: PatchMainContext,
    workspace_name: str,
    workspace_path: str,
    *,
    dry_run: bool,
    verbose: bool,
) -> None:
    ctx: AdoptContext = adopt_ctx(
        workspace=workspace_name,
        path=workspace_path,
        dry_run=dry_run,
        verbose=verbose,
    )

    run_adopt_mock = create_autospec(mainmod.run_adopt)
    monkeypatch.setattr(mainmod, "run_adopt", run_adopt_mock)
    patch_main_context(ctx=ctx, resolved=resolved_config)

    mainmod.main()

    run_adopt_mock.assert_called_once()
    call_kwargs = run_adopt_mock.call_args.kwargs
    assert "workspace" in call_kwargs
    assert call_kwargs["dry_run"] is dry_run
    assert call_kwargs["verbose"] is verbose


@pytest.mark.parametrize(
    ("init_path", "config_filename"),
    [
        (Path("/init"), DEFAULT_CONFIG_FILENAME),
    ],
)
def test_main_runs_init_success(
    monkeypatch: pytest.MonkeyPatch,
    config_path: Path,
    init_path: Path,
    config_filename: str,
) -> None:
    ctx = InitContext(
        command="init",
        config_path=config_path,
        config_filename=config_filename,
        path=init_path,
        dry_run=False,
        verbose=False,
    )

    parse_cli_mock = create_autospec(mainmod.parse_cli, return_value=ctx)
    run_init_mock = create_autospec(mainmod.run_init)

    monkeypatch.setattr(mainmod, "parse_cli", parse_cli_mock)
    monkeypatch.setattr(mainmod, "run_init", run_init_mock)

    mainmod.main()

    run_init_mock.assert_called_once()
    call_kwargs = run_init_mock.call_args.kwargs
    assert call_kwargs["path"] == ctx.path
    assert call_kwargs["config_filename"] == ctx.config_filename


@pytest.mark.parametrize(
    ("exc_type", "message", "init_path", "config_filename"),
    [
        (FileExistsError, "already exists", Path("/init"), DEFAULT_CONFIG_FILENAME),
        (mainmod.ConfigError, "bad config", Path("/init"), DEFAULT_CONFIG_FILENAME),
    ],
)
def test_main_exits_on_init_errors(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    config_path: Path,
    exc_type: type[Exception],
    message: str,
    init_path: Path,
    config_filename: str,
) -> None:
    ctx = InitContext(
        command="init",
        config_path=config_path,
        config_filename=config_filename,
        path=init_path,
        dry_run=False,
        verbose=False,
    )

    def raise_error(*_args: object, **_kwargs: object) -> None:
        raise exc_type(message)

    parse_cli_mock = create_autospec(mainmod.parse_cli, return_value=ctx)
    run_init_mock = create_autospec(mainmod.run_init, side_effect=raise_error)

    monkeypatch.setattr(mainmod, "parse_cli", parse_cli_mock)
    monkeypatch.setattr(mainmod, "run_init", run_init_mock)

    with pytest.raises(SystemExit) as excinfo:
        mainmod.main()

    assert excinfo.value.code == 1
    assert message in capsys.readouterr().err



import asyncio
import shutil
from typing import TYPE_CHECKING, TextIO

from contentctl.plan.sync import SyncAction, SyncOperation
from contentctl.utils.concurrent import map_concurrent

if TYPE_CHECKING:
    from collections.abc import AsyncIterable, AsyncIterator
    from pathlib import Path


async def print_sync_plan(
    operations: AsyncIterable[SyncOperation],
    destination_root: Path,
    output: TextIO,
) -> AsyncIterator[SyncOperation]:
    async for operation in operations:
        _print_operation(
            operation,
            destination_root,
            output,
        )
        yield operation


async def apply_sync_plan(
    operations: AsyncIterable[SyncOperation],
    source_root: Path,
    destination_root: Path,
    semaphore: asyncio.Semaphore,
) -> AsyncIterator[SyncOperation]:
    async def apply(operation: SyncOperation) -> SyncOperation:
        if operation.action is not SyncAction.SKIP:
            await asyncio.to_thread(
                _apply_operation,
                operation,
                source_root,
                destination_root,
            )
        return operation

    async for operation in map_concurrent(operations, apply, semaphore):
        yield operation


def _apply_operation(
    operation: SyncOperation,
    source_root: Path,
    destination_root: Path,
) -> None:
    destination = destination_root / operation.relative

    if operation.action is SyncAction.DELETE:
        if destination.exists():
            destination.unlink()
    elif operation.action in (SyncAction.COPY, SyncAction.REPLACE):
        source = source_root / operation.relative
        destination.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(source, destination)


def _print_operation(
    operation: SyncOperation,
    destination_root: Path,
    output: TextIO,
) -> None:
    destination = destination_root / operation.relative
    action_label = operation.action.value.lower()
    print(f"{action_label:<7} {destination}", file=output)



from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from contentctl.config import ResolvedConfig, select_workspaces

if TYPE_CHECKING:
    from tests.contentctl.fixture_types import MakeWorkspace


@pytest.mark.parametrize(
    ("selection_order", "expected_order"),
    [
        (["zeta", "alpha"], ["zeta", "alpha"]),
        (["alpha", "zeta"], ["alpha", "zeta"]),
        (["alpha"], ["alpha"]),
        (["zeta"], ["zeta"]),
    ],
)
def test_select_workspaces_preserves_order(
    make_workspace: MakeWorkspace,
    selection_order: list[str],
    expected_order: list[str],
) -> None:
    resolved = ResolvedConfig(
        origin=make_workspace("", Path("/origin")),
        workspaces={
            "alpha": make_workspace("alpha", Path("/alpha")),
            "zeta": make_workspace("zeta", Path("/zeta")),
        },
    )

    selected = select_workspaces(resolved, selection_order)

    assert [workspace.name for workspace in selected] == expected_order



from .loader import ConfigError, load_config
from .resolver import ResolvedConfig, Workspace, resolve_config
from .selectors import select_all_workspaces, select_workspaces

__all__ = [
    "ConfigError",
    "ResolvedConfig",
    "Workspace",
    "load_config",
    "resolve_config",
    "select_all_workspaces",
    "select_workspaces",
]



"""Fixtures for config module tests."""

from typing import TYPE_CHECKING, Any

import pytest
import yaml

if TYPE_CHECKING:
    from pathlib import Path

    from .fixture_types import LoadYamlFixture


@pytest.fixture
def load_yaml_fixture() -> LoadYamlFixture:
    """Load a YAML fixture file."""

    def _load(path: Path) -> dict[str, Any]:
        return yaml.safe_load(path.read_text(encoding="utf-8"))

    return _load



from pathlib import Path

import pytest
from contentctl.gateway import AdoptContext, DeployContext, InitContext, parse_cli
from contentctl.gateway.defaults import (
    DEFAULT_CONFIG_FILENAME,
    DEFAULT_CONFIG_PATH,
    DEFAULT_INIT_DIR,
)

VIRTUAL_WORKSPACE = "virtual-workspace"
VIRTUAL_SUBDIR = "virtual-subdir"
VIRTUAL_CONFIG_PATH = "virtual-config.yaml"


@pytest.mark.parametrize(
    (
        "argv",
        "ctx_type",
        "command",
        "path",
        "workspaces",
        "workspace",
        "config_name",
        "all_workspaces",
    ),
    [
        (
            [
                "--config",
                VIRTUAL_CONFIG_PATH,
                "deploy",
                VIRTUAL_WORKSPACE,
                "--path",
                VIRTUAL_SUBDIR,
            ],
            DeployContext,
            "deploy",
            VIRTUAL_SUBDIR,
            [VIRTUAL_WORKSPACE],
            None,
            VIRTUAL_CONFIG_PATH,
            False,
        ),
        (
            [
                "--config",
                VIRTUAL_CONFIG_PATH,
                "deploy",
                "--all-workspaces",
            ],
            DeployContext,
            "deploy",
            ".",
            [],
            None,
            VIRTUAL_CONFIG_PATH,
            True,
        ),
        (
            [
                "--config",
                VIRTUAL_CONFIG_PATH,
                "adopt",
                VIRTUAL_WORKSPACE,
            ],
            AdoptContext,
            "adopt",
            ".",
            None,
            VIRTUAL_WORKSPACE,
            VIRTUAL_CONFIG_PATH,
            None,
        ),
    ],
)
def test_parse_cli_success(
    fixture_dir: Path,
    argv: list[str],
    ctx_type: type[DeployContext | AdoptContext],
    command: str,
    path: str,
    workspaces: list[str] | None,
    workspace: str | None,
    config_name: str,
    *,
    all_workspaces: bool | None,
) -> None:
    ctx = parse_cli(argv, fixture_dir)

    assert isinstance(ctx, ctx_type)
    assert ctx.command == command
    assert ctx.path == path
    assert ctx.config_path == (fixture_dir / config_name).resolve()
    if isinstance(ctx, DeployContext):
        assert ctx.all_workspaces is all_workspaces
        assert ctx.workspaces == workspaces
    else:
        assert ctx.workspace == workspace


@pytest.mark.parametrize(
    "argv",
    [
        ["deploy", VIRTUAL_WORKSPACE, "--all-workspaces"],
        ["deploy"],
        ["deploy", ""],
        ["deploy", VIRTUAL_WORKSPACE, "--path", "   "],
        ["adopt", ""],
    ],
)
def test_parse_cli_rejects_invalid_args(fixture_dir: Path, argv: list[str]) -> None:
    with pytest.raises(SystemExit):
        parse_cli(argv, fixture_dir)


@pytest.mark.parametrize(
    ("config_arg", "workspace", "use_absolute"),
    [
        (VIRTUAL_CONFIG_PATH, VIRTUAL_WORKSPACE, False),
        ("virtual/custom/path/config.yaml", VIRTUAL_WORKSPACE, False),
        (VIRTUAL_CONFIG_PATH, VIRTUAL_WORKSPACE, True),
    ],
)
def test_parse_cli_accepts_config_path(
    fixture_dir: Path,
    config_arg: str,
    workspace: str,
    *,
    use_absolute: bool,
) -> None:
    config_path = fixture_dir / config_arg
    arg = str(config_path) if use_absolute else config_arg
    ctx = parse_cli(["--config", arg, "deploy", workspace], fixture_dir)

    if use_absolute:
        assert ctx.config_path == config_path.resolve()
    else:
        assert ctx.config_path == (fixture_dir / config_arg).resolve()


@pytest.mark.parametrize(
    (
        "pre_flags",
        "post_flags",
        "workspace",
        "expected_dry_run",
        "expected_verbose",
        "expected_delete",
    ),
    [
        (["--dry-run", "--verbose"], [], VIRTUAL_WORKSPACE, True, True, False),
        (["--dry-run"], [], VIRTUAL_WORKSPACE, True, False, False),
        (["--verbose"], [], VIRTUAL_WORKSPACE, False, True, False),
        ([], ["--delete"], VIRTUAL_WORKSPACE, False, False, True),
        ([], [], VIRTUAL_WORKSPACE, False, False, False),
        (["--dry-run"], ["--delete"], VIRTUAL_WORKSPACE, True, False, True),
    ],
)
def test_parse_cli_sets_flags(
    fixture_dir: Path,
    pre_flags: list[str],
    post_flags: list[str],
    workspace: str,
    *,
    expected_dry_run: bool,
    expected_verbose: bool,
    expected_delete: bool,
) -> None:
    ctx = parse_cli(
        [*pre_flags, "deploy", workspace, *post_flags],
        fixture_dir,
    )

    assert isinstance(ctx, DeployContext)
    assert ctx.dry_run is expected_dry_run
    assert ctx.verbose is expected_verbose
    assert ctx.allow_delete is expected_delete


@pytest.mark.parametrize(
    ("argv", "expected_path", "expected_is_absolute"),
    [
        (["init"], Path(DEFAULT_INIT_DIR), False),
        (["init", "virtual-subdir"], Path("virtual-subdir"), False),
        (["init", "/virtual/abs"], Path("/virtual/abs"), True),
    ],
)
def test_parse_cli_init_resolves_paths(
    fixture_dir: Path,
    argv: list[str],
    expected_path: Path,
    *,
    expected_is_absolute: bool,
) -> None:
    ctx = parse_cli(argv, fixture_dir)

    assert isinstance(ctx, InitContext)
    if expected_is_absolute:
        assert ctx.path == expected_path
    else:
        assert ctx.path == (fixture_dir / expected_path).resolve()
    assert ctx.config_path == (fixture_dir / DEFAULT_CONFIG_PATH).resolve()
    assert ctx.config_filename == DEFAULT_CONFIG_FILENAME


@pytest.mark.parametrize(
    ("flags", "expected_dry_run", "expected_verbose"),
    [
        (["--dry-run", "--verbose"], True, True),
        (["--dry-run"], True, False),
        (["--verbose"], False, True),
        ([], False, False),
    ],
)
def test_parse_cli_init_sets_flags(
    fixture_dir: Path,
    flags: list[str],
    *,
    expected_dry_run: bool,
    expected_verbose: bool,
) -> None:
    ctx = parse_cli([*flags, "init"], fixture_dir)

    assert isinstance(ctx, InitContext)
    assert ctx.dry_run is expected_dry_run
    assert ctx.verbose is expected_verbose



"""Config loading and validation for contentctl."""

from __future__ import annotations

import importlib.resources as resources
import json
import os
from collections import UserDict
from functools import lru_cache
from string import Template
from typing import TYPE_CHECKING, Any, Protocol, cast

import yaml
from jsonschema import Draft202012Validator

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from jsonschema.exceptions import ValidationError


def load_config(config_path: Path) -> dict[str, Any]:
    if not config_path.exists():
        raise ConfigError(f"config file not found: {config_path}")

    try:
        raw = config_path.read_text(encoding="utf-8")
    except OSError as exc:
        raise ConfigError(f"cannot read config file: {config_path}") from exc

    try:
        config = yaml.safe_load(raw)
    except yaml.YAMLError as exc:
        raise ConfigError(f"invalid yaml in config file: {config_path}") from exc

    match config:
        case None:
            raise ConfigError(f"config file is empty: {config_path}")
        case dict():
            pass
        case _:
            raise ConfigError("config root must be a mapping.")

    config = _render_env_vars(config)
    config_dict = cast("dict[str, Any]", config)
    _validate_schema(config_dict)
    return config_dict


class ConfigError(ValueError):
    pass


def _validate_schema(config: dict[str, Any]) -> None:
    schema = _load_schema()
    validator = Draft202012Validator(schema)
    errors = sorted(
        cast("_Validator", validator).iter_errors(config),
        key=lambda err: tuple(err.path),
    )
    if not errors:
        return

    details = "\n".join(f"- {_format_error_path(err)}: {err.message}" for err in errors)
    raise ConfigError(f"config schema validation failed:\n{details}")


def _render_env_vars(value: Any) -> Any:
    match value:
        case str():
            return Template(value).substitute(_EnvVars(os.environ))
        case list():
            value_list = cast("list[Any]", value)
            return [_render_env_vars(item) for item in value_list]
        case dict():
            rendered: dict[str, Any] = {}
            value_dict = cast("dict[str, Any]", value)
            for key, item in value_dict.items():
                rendered[key] = _render_env_vars(item)
            return rendered
        case _:
            return value


class _Validator(Protocol):
    def iter_errors(self, instance: Any) -> Iterable[ValidationError]: ...


def _format_error_path(error: ValidationError) -> str:
    if not error.path:
        return "<root>"
    segments: list[str] = []
    for segment in error.path:
        match segment:
            case int():
                segments.append(f"[{segment}]")
            case _:
                if segments:
                    segments.append(".")
                segments.append(str(segment))
    return "".join(segments)


class _EnvVars(UserDict[str, str]):
    def __missing__(self, key: str) -> str:
        return ""


@lru_cache
def _load_schema() -> dict[str, Any]:
    raw = resources.read_text(
        "contentctl.schema",
        "content-hub.schema.json",
    )
    return cast("dict[str, Any]", json.loads(raw))



import asyncio
import contextlib
import os
from dataclasses import dataclass
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from collections.abc import AsyncIterable, AsyncIterator, Awaitable, Callable

_DEFAULT_BUFFER = 256
_DEFAULT_CONCURRENCY = min(32, (os.cpu_count() or 1) + 4)


def default_concurrency() -> int:
    return _DEFAULT_CONCURRENCY


async def map_concurrent[Item, Result](
    source: AsyncIterable[Item],
    func: Callable[[Item], Awaitable[Result]],
    semaphore: asyncio.Semaphore,
) -> AsyncIterator[Result]:
    async def worker(item: Item) -> AsyncIterator[Emit[Result] | Spawn[Result]]:
        yield Emit(await func(item))

    async def main() -> AsyncIterator[Emit[Result] | Spawn[Result]]:
        async for item in source:
            yield Spawn(worker(item), semaphore)

    async for result in stream_concurrently(main()):
        yield result


async def stream_concurrently[Result](
    entry_point: AsyncIterator[Emit[Result] | Spawn[Result]],
    buffer: int = _DEFAULT_BUFFER,
) -> AsyncIterator[Result]:
    queue: asyncio.Queue[Result | _Done | _Error] = asyncio.Queue(maxsize=buffer)

    async def drive(
        iterator: AsyncIterator[Emit[Result] | Spawn[Result]],
        tg: asyncio.TaskGroup,
    ) -> None:
        async for op in iterator:
            match op:
                case Emit(value):
                    await queue.put(value)
                case Spawn(sub_iterator, semaphore):
                    if semaphore is None:
                        tg.create_task(drive(sub_iterator, tg))
                        continue

                    await semaphore.acquire()

                    async def drive_sem(
                        iterator: AsyncIterator[Emit[Result] | Spawn[Result]],
                        semaphore: asyncio.Semaphore,
                    ) -> None:
                        try:
                            await drive(iterator, tg)
                        finally:
                            semaphore.release()

                    try:
                        tg.create_task(drive_sem(sub_iterator, semaphore))
                    except:
                        semaphore.release()
                        raise

    async def producer() -> None:
        try:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(drive(entry_point, tg))

            await queue.put(_Done())
        except asyncio.CancelledError:
            raise
        except BaseException as exc:
            await queue.put(_Error(_normalize_exception(exc)))

    task = asyncio.create_task(producer())

    try:
        while True:
            item = await queue.get()

            match item:
                case _Done():
                    break
                case _Error(exc):
                    raise exc
                case _:
                    yield item

    finally:
        if not task.done():
            task.cancel()

        with contextlib.suppress(asyncio.CancelledError):
            await task


@dataclass(slots=True)
class Emit[T]:
    value: T


@dataclass(slots=True)
class Spawn[T]:
    iterator: AsyncIterator[Emit[T] | Spawn[T]]
    semaphore: asyncio.Semaphore | None = None


@dataclass(slots=True)
class _Done:
    pass


@dataclass(slots=True)
class _Error:
    exc: BaseException


def _normalize_exception(exc: BaseException) -> BaseException:
    if not isinstance(exc, ExceptionGroup):
        return exc

    group = cast("ExceptionGroup[Exception]", exc)
    exceptions = group.exceptions
    if len(exceptions) != 1:
        return cast("BaseException", exc)

    return _normalize_exception(exceptions[0])



"""Fixtures specific to contentctl module tests."""

from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from contentctl.config import Workspace
from contentctl.plan.sync import SyncAction, SyncOperation

if TYPE_CHECKING:
    from .fixture_types import MakeSyncOp, MakeWorkspace


@pytest.fixture
def make_workspace() -> MakeWorkspace:
    """Create a Workspace instance for testing."""

    def _make(name: str, root: str | Path) -> Workspace:
        return Workspace(name=name, path=Path(root), include=(), exclude=())

    return _make


@pytest.fixture
def make_sync_op() -> MakeSyncOp:
    """Create a SyncOperation for testing."""

    def _make(filename: str, action: SyncAction) -> SyncOperation:
        return SyncOperation(relative=Path(filename), action=action)

    return _make



import asyncio
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, TextIO, cast
from unittest.mock import create_autospec

import pytest
from contentctl.operations import adopt as adopt_mod
from contentctl.operations import deploy as deploy_mod
from contentctl.operations.adopt import run_adopt
from contentctl.operations.deploy import run_deploy
from contentctl.plan import SyncPlan
from contentctl.plan.sync import SyncAction, SyncOperation

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from tests.contentctl.fixture_types import MakeWorkspace
    from tests.fixture_types import FixturePath


@pytest.mark.parametrize(
    (
        "source",
        "dest",
        "workspace_name",
        "workspace_root",
        "origin_root",
        "sync_path",
        "planned_filename",
        "dry_run",
        "verbose",
    ),
    [
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            "docs",
            "/ws",
            "/origin",
            ".",
            "guide.txt",
            True,
            True,
        ),
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            "docs",
            "/ws",
            "/origin",
            ".",
            "guide.txt",
            True,
            False,
        ),
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            "docs",
            "/ws",
            "/origin",
            ".",
            "guide.txt",
            False,
            True,
        ),
        (
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            "docs",
            "/ws",
            "/origin",
            ".",
            "guide.txt",
            False,
            False,
        ),
    ],
)
def test_run_adopt_applies_plan(
    monkeypatch: pytest.MonkeyPatch,
    fixture_path: FixturePath,
    make_workspace: MakeWorkspace,
    source: tuple[str, str],
    dest: tuple[str, str],
    workspace_name: str,
    workspace_root: str,
    origin_root: str,
    sync_path: str,
    planned_filename: str,
    *,
    dry_run: bool,
    verbose: bool,
) -> None:
    execute_calls: list[tuple[bool, bool]] = []

    async def observe_execute(*_args: object, **kwargs: object) -> int:
        execute_calls.append((bool(kwargs.get("dry_run")), bool(kwargs.get("verbose"))))
        if output := kwargs.get("output"):
            operation_name = kwargs.get("operation_name", "")
            workspace_name = kwargs.get("workspace_name", "")
            print(
                f"{operation_name} {workspace_name}: 1 files copied",
                file=cast("TextIO", output),
            )
        return 1

    execute_mock = create_autospec(
        adopt_mod.execute_sync_operation,
        side_effect=observe_execute,
    )

    def plan_sync_side_effect(
        *_args: object,
        **_kwargs: object,
    ) -> SyncPlan:
        async def iter_ops() -> AsyncIterator[SyncOperation]:
            yield SyncOperation(
                relative=Path(planned_filename),
                action=SyncAction.COPY,
            )

        source_path = fixture_path(*source)
        destination_path = fixture_path(*dest)
        return SyncPlan(
            stream=iter_ops(),
            source_path=source_path,
            destination_path=destination_path,
            source_root=source_path,
            destination_root=destination_path,
        )

    plan_sync_mock = create_autospec(
        adopt_mod.plan_sync,
        side_effect=plan_sync_side_effect,
    )
    monkeypatch.setattr(adopt_mod, "plan_sync", plan_sync_mock)
    monkeypatch.setattr(adopt_mod, "execute_sync_operation", execute_mock)

    output = StringIO()
    asyncio.run(
        run_adopt(
            workspace=make_workspace(workspace_name, workspace_root),
            origin=make_workspace("", origin_root),
            path=sync_path,
            dry_run=dry_run,
            verbose=verbose,
            output=output,
        )
    )

    text = output.getvalue()
    assert f"adopt {workspace_name}:" in text
    assert "1 files copied" in text
    assert execute_calls == [(dry_run, verbose)]


@pytest.mark.parametrize(
    (
        "workspace_names",
        "source_fixture",
        "destination_fixture",
        "workspace_root_prefix",
        "origin_root",
        "sync_path",
        "planned_filename",
        "resolved_source_root",
        "resolved_destination_root",
        "dry_run",
        "verbose",
        "allow_delete",
    ),
    [
        (
            ("docs", "assets"),
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            "/",
            "/origin",
            ".",
            "guide.txt",
            "/source",
            "/destination",
            True,
            True,
            True,
        ),
        (
            ("docs", "assets"),
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            "/",
            "/origin",
            ".",
            "guide.txt",
            "/source",
            "/destination",
            True,
            False,
            False,
        ),
        (
            ("docs", "assets"),
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            "/",
            "/origin",
            ".",
            "guide.txt",
            "/source",
            "/destination",
            False,
            True,
            False,
        ),
        (
            ("docs", "assets"),
            ("plan_sync", "source_multi"),
            ("plan_sync", "destination_single"),
            "/",
            "/origin",
            ".",
            "guide.txt",
            "/source",
            "/destination",
            False,
            False,
            True,
        ),
    ],
)
def test_run_deploy_applies_plan(
    monkeypatch: pytest.MonkeyPatch,
    fixture_path: FixturePath,
    make_workspace: MakeWorkspace,
    workspace_names: tuple[str, ...],
    source_fixture: tuple[str, ...],
    destination_fixture: tuple[str, ...],
    workspace_root_prefix: str,
    origin_root: str,
    sync_path: str,
    planned_filename: str,
    resolved_source_root: str,
    resolved_destination_root: str,
    *,
    dry_run: bool,
    verbose: bool,
    allow_delete: bool,
) -> None:
    execute_calls: list[tuple[str, bool, bool]] = []
    allow_delete_calls: list[bool] = []

    async def observe_execute(*_args: object, **kwargs: object) -> int:
        execute_calls.append(
            (
                str(kwargs.get("workspace_name", "")),
                bool(kwargs.get("dry_run")),
                bool(kwargs.get("verbose")),
            )
        )
        return 1

    def plan_sync_side_effect(
        *_args: object,
        **kwargs: object,
    ) -> SyncPlan:
        policy = kwargs.get("policy")
        allow_delete_calls.append(bool(getattr(policy, "allow_delete", False)))

        async def iter_ops() -> AsyncIterator[SyncOperation]:
            yield SyncOperation(
                relative=Path(planned_filename),
                action=SyncAction.COPY,
            )

        return SyncPlan(
            stream=iter_ops(),
            source_path=fixture_path(*source_fixture),
            destination_path=fixture_path(*destination_fixture),
            source_root=Path(resolved_source_root),
            destination_root=Path(resolved_destination_root),
        )

    plan_sync_mock = create_autospec(
        deploy_mod.plan_sync,
        side_effect=plan_sync_side_effect,
    )
    execute_mock = create_autospec(
        deploy_mod.execute_sync_operation,
        side_effect=observe_execute,
    )

    monkeypatch.setattr(deploy_mod, "plan_sync", plan_sync_mock)
    monkeypatch.setattr(deploy_mod, "execute_sync_operation", execute_mock)

    workspaces = [
        make_workspace(name, f"{workspace_root_prefix}{name}")
        for name in workspace_names
    ]
    origin = make_workspace("", origin_root)

    output = StringIO()
    asyncio.run(
        run_deploy(
            workspaces=workspaces,
            origin=origin,
            path=sync_path,
            dry_run=dry_run,
            verbose=verbose,
            allow_delete=allow_delete,
            output=output,
        )
    )

    text = output.getvalue()
    if verbose or dry_run:
        assert all(f"deploy {name}:" in text for name in workspace_names)
    else:
        assert text == ""

    assert execute_calls == [(name, dry_run, verbose) for name in workspace_names]
    assert allow_delete_calls == [allow_delete, allow_delete]



from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import create_autospec

import pytest
from contentctl.gateway.defaults import DEFAULT_CONFIG_FILENAME
from contentctl.operations.init import run_init

if TYPE_CHECKING:
    from tests.fixture_types import FixturePath


@pytest.mark.parametrize(
    ("init_dir_name", "config_filename"),
    [("virtual-init", DEFAULT_CONFIG_FILENAME)],
)
def test_init_dry_run_skips_write(
    fixture_path: FixturePath,
    monkeypatch: pytest.MonkeyPatch,
    init_dir_name: str,
    config_filename: str,
) -> None:
    observed_writes: list[str] = []
    init_dir = fixture_path(init_dir_name)

    def observe_write_text(_self: Path, content: str, **_kwargs: object) -> None:
        observed_writes.append(content)

    write_text_mock = create_autospec(Path.write_text, side_effect=observe_write_text)
    monkeypatch.setattr(Path, "write_text", write_text_mock)

    output = StringIO()
    run_init(
        path=init_dir,
        config_filename=config_filename,
        dry_run=True,
        verbose=False,
        output=output,
    )

    write_text_mock.assert_not_called()
    assert observed_writes == []


@pytest.mark.parametrize(
    ("init_dir_name", "config_filename"),
    [("virtual-init", DEFAULT_CONFIG_FILENAME)],
)
def test_init_fails_when_file_exists(
    fixture_path: FixturePath,
    monkeypatch: pytest.MonkeyPatch,
    init_dir_name: str,
    config_filename: str,
) -> None:
    init_dir = fixture_path(init_dir_name)
    exists_mock = create_autospec(Path.exists, return_value=True)
    monkeypatch.setattr(Path, "exists", exists_mock)

    output = StringIO()
    with pytest.raises(FileExistsError):
        run_init(
            path=init_dir,
            config_filename=config_filename,
            dry_run=False,
            verbose=False,
            output=output,
        )


@pytest.mark.parametrize(
    ("init_dir_name", "config_filename"),
    [("virtual-init", DEFAULT_CONFIG_FILENAME)],
)
def test_init_writes_config_and_creates_directories(
    fixture_path: FixturePath,
    monkeypatch: pytest.MonkeyPatch,
    init_dir_name: str,
    config_filename: str,
) -> None:
    observed_writes: list[str] = []
    init_dir = fixture_path(init_dir_name)

    def observe_write_text(_self: Path, content: str, **_kwargs: object) -> None:
        observed_writes.append(content)

    mkdir_mock = create_autospec(Path.mkdir)
    write_text_mock = create_autospec(Path.write_text, side_effect=observe_write_text)

    monkeypatch.setattr(Path, "mkdir", mkdir_mock)
    monkeypatch.setattr(Path, "write_text", write_text_mock)

    output = StringIO()
    run_init(
        path=init_dir,
        config_filename=config_filename,
        dry_run=False,
        verbose=False,
        output=output,
    )

    mkdir_mock.assert_called_once()
    call_kwargs = mkdir_mock.call_args.kwargs
    assert call_kwargs["parents"] is True
    assert call_kwargs["exist_ok"] is True
    write_text_mock.assert_called_once()
    assert len(observed_writes) == 1
    assert observed_writes[0]



from __future__ import annotations

from typing import TYPE_CHECKING

from .loader import ConfigError

if TYPE_CHECKING:
    from collections.abc import Iterable

    from .resolver import ResolvedConfig, Workspace


def select_all_workspaces(resolved: ResolvedConfig) -> list[Workspace]:
    return [resolved.workspaces[name] for name in sorted(resolved.workspaces.keys())]


def select_workspaces(
    resolved: ResolvedConfig, workspaces: Iterable[str]
) -> list[Workspace]:
    selected: list[Workspace] = []
    for name in workspaces:
        try:
            selected.append(resolved.workspaces[name])
        except KeyError as exc:
            raise ConfigError(f"unknown workspace: {name}") from exc
    return selected



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, TextIO

from contentctl.operation_kit import execute_sync_operation
from contentctl.plan.sync import SyncFilters, SyncPolicy, SyncScope, plan_sync
from contentctl.utils.concurrent import default_concurrency

if TYPE_CHECKING:
    from contentctl.config import Workspace


async def run_adopt(
    workspace: Workspace,
    origin: Workspace,
    path: str,
    output: TextIO,
    *,
    dry_run: bool,
    verbose: bool,
) -> None:
    base = default_concurrency()
    io_semaphore = asyncio.Semaphore(base)

    plan = plan_sync(
        SyncScope(workspace.path, origin.path, path),
        SyncFilters(
            source_include=workspace.include,
            source_exclude=workspace.exclude,
            destination_include=origin.include,
            destination_exclude=origin.exclude,
        ),
        policy=SyncPolicy(semaphore=io_semaphore),
    )

    if verbose or dry_run:
        print(
            f"adopt {workspace.name}: {plan.source_path} -> {plan.destination_path}",
            file=output,
        )

    await execute_sync_operation(
        stream=plan.stream,
        source_root=plan.source_root,
        destination_root=plan.destination_root,
        semaphore=io_semaphore,
        operation_name="adopt",
        workspace_name=workspace.name,
        dry_run=dry_run,
        verbose=verbose,
        output=output,
    )



import asyncio
import sys
from pathlib import Path

from .config import (
    ConfigError,
    ResolvedConfig,
    load_config,
    resolve_config,
    select_all_workspaces,
    select_workspaces,
)
from .gateway import AdoptContext, DeployContext, InitContext, parse_cli
from .operations import SyncError, run_adopt, run_deploy, run_init


def main() -> None:
    ctx = parse_cli(sys.argv[1:], Path.cwd())

    # Init command doesn't require existing config
    if isinstance(ctx, InitContext):
        try:
            run_init(
                path=ctx.path,
                config_filename=ctx.config_filename,
                dry_run=ctx.dry_run,
                verbose=ctx.verbose,
                output=sys.stdout,
            )
        except (ConfigError, FileExistsError) as exc:
            print(f"error: {exc}", file=sys.stderr)
            sys.exit(1)
        except PermissionError as exc:
            _print_permission_error(exc)
            sys.exit(1)
        return

    try:
        raw_config = load_config(ctx.config_path)
        resolved = resolve_config(raw_config, ctx.config_path)
    except ConfigError as exc:
        print(f"error: {exc}", file=sys.stderr)
        sys.exit(2)

    try:
        asyncio.run(_dispatch(ctx, resolved))
    except (ConfigError, SyncError) as exc:
        print(f"error: {exc}", file=sys.stderr)
        sys.exit(1)
    except PermissionError as exc:
        _print_permission_error(exc)
        sys.exit(1)


async def _dispatch(
    ctx: AdoptContext | DeployContext,
    resolved: ResolvedConfig,
) -> None:
    match ctx:
        case DeployContext():
            if ctx.all_workspaces:
                workspaces = select_all_workspaces(resolved)
            else:
                workspaces = select_workspaces(resolved, ctx.workspaces)
            await run_deploy(
                workspaces=workspaces,
                origin=resolved.origin,
                path=ctx.path,
                dry_run=ctx.dry_run,
                verbose=ctx.verbose,
                allow_delete=ctx.allow_delete,
                output=sys.stdout,
            )
        case AdoptContext():
            workspaces = select_workspaces(resolved, [ctx.workspace])
            await run_adopt(
                workspace=workspaces[0],
                origin=resolved.origin,
                path=ctx.path,
                dry_run=ctx.dry_run,
                verbose=ctx.verbose,
                output=sys.stdout,
            )


def _print_permission_error(exc: PermissionError) -> None:
    if exc.filename:
        print(f"error: permission denied: {exc.filename}", file=sys.stderr)
    else:
        print("error: permission denied", file=sys.stderr)


if __name__ == "__main__":
    main()



"""CLI parsing and context resolution for contentctl."""

import argparse
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Literal

from .defaults import (
    DEFAULT_CONFIG_FILENAME,
    DEFAULT_CONFIG_PATH,
    DEFAULT_INIT_DIR,
    DEFAULT_WORKSPACE_SUBPATH,
)

if TYPE_CHECKING:
    from collections.abc import Callable


def parse_cli(argv: list[str], cwd: Path) -> CliContext:
    parser = _build_parser()
    args = parser.parse_args(argv)

    config_path = _resolve_config_path(cwd, args.config)

    if not args.path.strip():
        parser.error("path is required.")

    match args.command:
        case "deploy":
            _validate_relative_path(parser, args.path)
            workspaces = list(args.workspace or [])
            all_workspaces = bool(args.all_workspaces)
            _validate_deploy_workspaces(
                parser,
                workspaces,
                all_workspaces=all_workspaces,
            )
            return DeployContext(
                command="deploy",
                config_path=config_path,
                all_workspaces=all_workspaces,
                workspaces=workspaces,
                path=args.path,
                allow_delete=args.delete,
                dry_run=args.dry_run,
                verbose=args.verbose,
            )
        case "adopt":
            _validate_relative_path(parser, args.path)
            workspace = args.workspace
            _validate_adopt_workspace(parser, workspace)
            return AdoptContext(
                command="adopt",
                config_path=config_path,
                workspace=workspace,
                path=args.path,
                dry_run=args.dry_run,
                verbose=args.verbose,
            )
        case "init":
            init_path = Path(args.path)
            if not init_path.is_absolute():
                init_path = cwd / init_path
            return InitContext(
                command="init",
                config_path=config_path,
                config_filename=DEFAULT_CONFIG_FILENAME,
                path=init_path.resolve(),
                dry_run=args.dry_run,
                verbose=args.verbose,
            )
        case _:
            parser.error(f"unknown command: {args.command}")


@dataclass(frozen=True)
class BaseContext:
    config_path: Path
    dry_run: bool
    verbose: bool


@dataclass(frozen=True)
class DeployContext(BaseContext):
    command: Literal["deploy"]
    all_workspaces: bool
    workspaces: list[str]
    path: str
    allow_delete: bool


@dataclass(frozen=True)
class AdoptContext(BaseContext):
    command: Literal["adopt"]
    workspace: str
    path: str


@dataclass(frozen=True)
class InitContext(BaseContext):
    command: Literal["init"]
    path: Path
    config_filename: str


CliContext = DeployContext | AdoptContext | InitContext


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="contentctl",
        description="sync directory content between an origin and workspaces.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-c",
        "--config",
        dest="config",
        default=DEFAULT_CONFIG_PATH,
        help="path to the content-hub config file.",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="show planned operations without writing changes.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="show detailed operation output.",
    )
    subparsers = parser.add_subparsers(dest="command", required=True)
    _add_deploy_parser(subparsers.add_parser)
    _add_adopt_parser(subparsers.add_parser)
    _add_init_parser(subparsers.add_parser)
    return parser


def _resolve_config_path(cwd: Path, config_arg: str) -> Path:
    config_path = Path(config_arg)
    if not config_path.is_absolute():
        config_path = cwd / config_path
    return config_path.resolve()


def _validate_relative_path(
    parser: argparse.ArgumentParser,
    path: str,
) -> None:
    if Path(path).is_absolute():
        parser.error("path must be relative.")


def _validate_deploy_workspaces(
    parser: argparse.ArgumentParser,
    workspaces: list[str],
    *,
    all_workspaces: bool,
) -> None:
    if all_workspaces and workspaces:
        parser.error("use workspace list or --all-workspaces, not both.")
    if not all_workspaces and not workspaces:
        parser.error("workspace is required unless --all-workspaces is set.")
    for workspace in workspaces:
        if not workspace.strip():
            parser.error("workspace cannot be empty.")


def _validate_adopt_workspace(
    parser: argparse.ArgumentParser,
    workspace: str,
) -> None:
    if not workspace.strip():
        parser.error("workspace cannot be empty.")


def _add_deploy_parser(
    add_parser: Callable[..., argparse.ArgumentParser],
) -> None:
    parser = add_parser(
        "deploy",
        help="deploy content from origin to workspace.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--all-workspaces",
        action="store_true",
        help="target all workspaces in the config.",
    )
    parser.add_argument(
        "--delete",
        action="store_true",
        help="delete workspace files missing in origin.",
    )
    _add_workspace_arg(parser, nargs="*")
    _add_path_arg(parser)


def _add_adopt_parser(
    add_parser: Callable[..., argparse.ArgumentParser],
) -> None:
    parser = add_parser(
        "adopt",
        help="adopt content from workspace to origin.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    _add_workspace_arg(parser)
    _add_path_arg(parser)


def _add_init_parser(
    add_parser: Callable[..., argparse.ArgumentParser],
) -> None:
    parser = add_parser(
        "init",
        help=f"create a new {DEFAULT_CONFIG_FILENAME} config file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "path",
        nargs="?",
        default=DEFAULT_INIT_DIR,
        help=f"target directory for {DEFAULT_CONFIG_FILENAME}.",
    )


def _add_workspace_arg(
    parser: argparse.ArgumentParser,
    nargs: str | None = None,
) -> None:
    parser.add_argument(
        "workspace",
        nargs=nargs,
        help="workspace alias from the config.",
    )


def _add_path_arg(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-p",
        "--path",
        default=DEFAULT_WORKSPACE_SUBPATH,
        help="relative subpath under origin and workspace roots.",
    )



from typing import TYPE_CHECKING

import pytest
from contentctl.config import (
    ConfigError,
    resolve_config,
    select_all_workspaces,
    select_workspaces,
)

if TYPE_CHECKING:
    from tests.fixture_types import FixturePath

    from .fixture_types import LoadYamlFixture


@pytest.mark.parametrize(
    ("config_file", "expected_origin_include", "expected_origin_exclude"),
    [
        ("resolver_defaults.yaml", ("src/**", "docs/**"), ("build/**",)),
        (
            "resolver_patterns.yaml",
            ("docs/**", "assets/**", "origin/**"),
            ("build/**",),
        ),
    ],
)
def test_resolve_config_origin_patterns(
    fixture_path: FixturePath,
    load_yaml_fixture: LoadYamlFixture,
    config_file: str,
    expected_origin_include: tuple[str, ...],
    expected_origin_exclude: tuple[str, ...],
) -> None:
    config_path = fixture_path(config_file)
    config = load_yaml_fixture(config_path)
    base_dir = config_path.parent

    resolved = resolve_config(config, config_path)

    assert resolved.origin.path == (base_dir / "origin").resolve()
    assert resolved.origin.include == expected_origin_include
    assert resolved.origin.exclude == expected_origin_exclude


@pytest.mark.parametrize(
    ("config_file", "workspace_name", "expected_include", "expected_exclude"),
    [
        (
            "resolver_defaults.yaml",
            "docs",
            ("src/**", "docs/**", "api/**"),
            ("build/**", "drafts/**"),
        ),
        (
            "resolver_defaults.yaml",
            "assets",
            ("src/**", "docs/**"),
            ("build/**",),
        ),
        (
            "resolver_patterns.yaml",
            "site",
            ("docs/**", "assets/**", "site/**"),
            ("build/**", "site/tmp/**"),
        ),
        (
            "resolver_patterns.yaml",
            "assets",
            ("docs/**", "assets/**"),
            ("build/**",),
        ),
    ],
)
def test_resolve_config_workspace_patterns(
    fixture_path: FixturePath,
    load_yaml_fixture: LoadYamlFixture,
    config_file: str,
    workspace_name: str,
    expected_include: tuple[str, ...],
    expected_exclude: tuple[str, ...],
) -> None:
    config_path = fixture_path(config_file)
    config = load_yaml_fixture(config_path)
    base_dir = config_path.parent

    resolved = resolve_config(config, config_path)

    workspace = resolved.workspaces[workspace_name]
    assert workspace.path == (base_dir / workspace_name).resolve()
    assert workspace.include == expected_include
    assert workspace.exclude == expected_exclude


@pytest.mark.parametrize(
    ("config_file", "expected_names"),
    [
        ("resolver_sort.yaml", ["alpha", "zeta"]),
    ],
)
def test_select_all_workspaces_sorted(
    fixture_path: FixturePath,
    load_yaml_fixture: LoadYamlFixture,
    config_file: str,
    expected_names: list[str],
) -> None:
    config_path = fixture_path(config_file)
    config = load_yaml_fixture(config_path)

    resolved = resolve_config(config, config_path)
    names = [workspace.name for workspace in select_all_workspaces(resolved)]

    assert names == expected_names


@pytest.mark.parametrize(
    "config_file",
    ["resolver_unknown.yaml"],
)
def test_select_workspaces_unknown(
    fixture_path: FixturePath, load_yaml_fixture: LoadYamlFixture, config_file: str
) -> None:
    config_path = fixture_path(config_file)
    config = load_yaml_fixture(config_path)
    resolved = resolve_config(config, config_path)

    with pytest.raises(ConfigError):
        select_workspaces(resolved, ["missing"])



from __future__ import annotations

from typing import TYPE_CHECKING, TextIO

from contentctl.execute.sync import apply_sync_plan, print_sync_plan
from contentctl.plan.sync import SyncAction, SyncOperation
from contentctl.utils.streams import count_stream

if TYPE_CHECKING:
    import asyncio
    from collections.abc import AsyncIterable
    from pathlib import Path


async def execute_sync_operation(
    stream: AsyncIterable[SyncOperation],
    source_root: Path,
    destination_root: Path,
    semaphore: asyncio.Semaphore,
    operation_name: str,
    workspace_name: str,
    output: TextIO,
    *,
    dry_run: bool,
    verbose: bool,
) -> int:
    if dry_run:
        return await _count_planned_operations(
            stream,
            destination_root,
            operation_name,
            workspace_name,
            output,
        )
    else:
        return await _execute_and_count_copied(
            stream,
            source_root,
            destination_root,
            semaphore,
            operation_name,
            workspace_name,
            output,
            verbose=verbose,
        )


async def _count_planned_operations(
    stream: AsyncIterable[SyncOperation],
    destination_root: Path,
    operation_name: str,
    workspace_name: str,
    output: TextIO,
) -> int:
    stream = print_sync_plan(stream, destination_root, output)
    total = await count_stream(stream)
    print(
        f"{operation_name} {workspace_name}: planned {total} files",
        file=output,
    )
    return total


async def _execute_and_count_copied(
    stream: AsyncIterable[SyncOperation],
    source_root: Path,
    destination_root: Path,
    semaphore: asyncio.Semaphore,
    operation_name: str,
    workspace_name: str,
    output: TextIO,
    *,
    verbose: bool,
) -> int:
    if verbose:
        stream = print_sync_plan(stream, destination_root, output)

    stream = apply_sync_plan(stream, source_root, destination_root, semaphore)
    total = await count_stream(
        stream, predicate=lambda op: op.action not in (SyncAction.SKIP,)
    )
    print(
        f"{operation_name} {workspace_name}: synced {total} files",
        file=output,
    )
    return total



from __future__ import annotations

from typing import TYPE_CHECKING, TextIO

import yaml

if TYPE_CHECKING:
    from pathlib import Path


def run_init(
    path: Path,
    config_filename: str,
    output: TextIO,
    *,
    dry_run: bool,
    verbose: bool,
) -> None:
    target_dir = path
    config_file = target_dir / config_filename

    if config_file.exists():
        raise FileExistsError(
            f"config file already exists: {config_file}\n"
            "remove it first or use a different path."
        )

    config_content = _generate_config_template()

    if verbose or dry_run:
        print(f"init: creating config file at {config_file}", file=output)

    if dry_run:
        print("init: config content:", file=output)
        print(config_content, file=output)
        return

    # Ensure directory exists
    target_dir.mkdir(parents=True, exist_ok=True)

    # Write config file
    config_file.write_text(config_content, encoding="utf-8")

    print(f"init: created {config_file}", file=output)


def _generate_config_template() -> str:
    config = {
        "origin": "origin",
        "workspaces": {
            "docs": "./docs",
        },
    }

    return yaml.dump(
        config,
        default_flow_style=False,
        sort_keys=False,
        allow_unicode=True,
    )



from typing import TYPE_CHECKING, Any

import pytest
from contentctl.config import ConfigError, load_config

if TYPE_CHECKING:
    from tests.fixture_types import FixturePath


@pytest.mark.parametrize(
    "config_file",
    ["config_valid.yaml"],
)
def test_load_config_valid(fixture_path: FixturePath, config_file: str) -> None:
    config_path = fixture_path(config_file)

    config = load_config(config_path)

    assert config["origin"] == "./origin"
    assert config["workspaces"]["docs"] == "./docs"


@pytest.mark.parametrize(
    ("config_file", "env_vars", "env_to_unset", "expected"),
    [
        (
            "config_env.yaml",
            {"SPACE_STATION": "/test/station"},
            [],
            {
                "origin": "/test/station/origin",
                "workspaces": {"docs": "/test/station/docs"},
            },
        ),
        (
            "config_env_nested.yaml",
            {"ROCKET_LAUNCHPAD": "/test/launchpad"},
            ["MISSING_VAR"],
            {
                "defaults": {"exclude": ["/test/launchpad/tmp/*"]},
                "origin": {
                    "path": "/test/launchpad/origin",
                    "include": ["/test/launchpad/docs/*.md", "/fallback/*.md"],
                },
                "workspaces": {
                    "docs": {
                        "path": "/test/launchpad/docs",
                        "include": ["/test/launchpad/docs/*.md"],
                        "exclude": ["/test/launchpad/docs/drafts/*.md"],
                    },
                    "assets": "/test/launchpad/assets",
                },
            },
        ),
    ],
)
def test_load_config_env_substitution(
    fixture_path: FixturePath,
    monkeypatch: pytest.MonkeyPatch,
    config_file: str,
    env_vars: dict[str, str],
    env_to_unset: list[str],
    expected: dict[str, Any],
) -> None:
    for key, value in env_vars.items():
        monkeypatch.setenv(key, value)
    for env_var in env_to_unset:
        monkeypatch.delenv(env_var, raising=False)
    config_path = fixture_path(config_file)

    config = load_config(config_path)

    assert config == expected


@pytest.mark.parametrize(
    "config_file",
    ["missing.yaml"],
)
def test_load_config_rejects_missing_file(
    fixture_path: FixturePath, config_file: str
) -> None:
    config_path = fixture_path(config_file)

    with pytest.raises(ConfigError):
        load_config(config_path)


@pytest.mark.parametrize(
    "config_file",
    [
        "config_empty.yaml",
        "config_invalid_yaml.yaml",
        "config_invalid_root.yaml",
        "config_schema_error.yaml",
    ],
)
def test_load_config_invalid_files(fixture_path: FixturePath, config_file: str) -> None:
    config_path = fixture_path(config_file)

    with pytest.raises(ConfigError):
        load_config(config_path)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import AsyncIterable, Callable


async def count_stream[StreamItem](
    stream: AsyncIterable[StreamItem],
    predicate: Callable[[StreamItem], bool] | None = None,
) -> int:
    count = 0
    if predicate is None:
        async for _ in stream:
            count += 1
        return count
    async for item in stream:
        if predicate(item):
            count += 1
    return count



from __future__ import annotations

import asyncio
import glob
import os
import re
from dataclasses import dataclass
from enum import StrEnum
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING

from contentctl.utils.concurrent import Emit, Spawn, stream_concurrently

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterator


def plan_sync(
    scope: SyncScope,
    filters: SyncFilters,
    *,
    policy: SyncPolicy,
) -> SyncPlan:
    _validate_subpath(scope.path)
    source_path = scope.source_root / scope.path
    destination_path = scope.destination_root / scope.path
    _validate_sync_paths(source_path, destination_path)

    source_is_file = source_path.is_file()
    source_plan_root = source_path.parent if source_is_file else source_path
    destination_plan_root = (
        destination_path.parent if source_is_file else destination_path
    )

    source_base = source_path.relative_to(scope.source_root)
    destination_base = destination_path.relative_to(scope.destination_root)
    if source_is_file:
        source_base = source_base.parent
        destination_base = destination_base.parent
    base = SyncBase(source=source_base, destination=destination_base)

    async def stream() -> AsyncIterator[SyncOperation]:
        if source_is_file:
            operation = _plan_sync_single_file(
                source_path,
                destination_path,
                filters=filters,
                base=base,
            )
            if operation:
                yield operation
            return

        async for operation in _plan_sync_directories(
            source_path,
            destination_path,
            filters=filters,
            base=base,
            policy=policy,
        ):
            yield operation

    return SyncPlan(
        stream=stream(),
        source_path=source_path,
        destination_path=destination_path,
        source_root=source_plan_root,
        destination_root=destination_plan_root,
    )


class SyncError(RuntimeError):
    pass


@dataclass(frozen=True)
class SyncOperation:
    relative: Path
    action: SyncAction


@dataclass(frozen=True)
class SyncScope:
    source_root: Path
    destination_root: Path
    path: str


@dataclass(frozen=True)
class SyncFilters:
    source_include: tuple[str, ...]
    source_exclude: tuple[str, ...]
    destination_include: tuple[str, ...]
    destination_exclude: tuple[str, ...]


@dataclass(frozen=True)
class SyncBase:
    source: Path
    destination: Path


@dataclass(frozen=True)
class SyncPolicy:
    semaphore: asyncio.Semaphore
    allow_delete: bool = False


@dataclass(frozen=True)
class SyncPlan:
    stream: AsyncIterator[SyncOperation]
    source_path: Path
    destination_path: Path
    source_root: Path
    destination_root: Path


class SyncAction(StrEnum):
    COPY = "COPY"
    REPLACE = "REPLACE"
    SKIP = "SKIP"
    DELETE = "DELETE"


@dataclass(frozen=True)
class _DirectoryEntries:
    files: set[Path]
    subdirs: set[Path]


def _validate_subpath(subpath: str) -> None:
    candidate = Path(subpath)
    if ".." in candidate.parts:
        raise SyncError(f"path escapes base directory: {subpath}")


def _plan_sync_single_file(
    source_path: Path,
    destination_path: Path,
    *,
    filters: SyncFilters,
    base: SyncBase,
) -> SyncOperation | None:
    rel_path = Path(source_path.name)

    if not _is_managed_with_base(
        rel_path, base.source, filters.source_include, filters.source_exclude
    ):
        return None

    if not _is_managed_with_base(
        rel_path,
        base.destination,
        filters.destination_include,
        filters.destination_exclude,
    ):
        return SyncOperation(relative=rel_path, action=SyncAction.SKIP)

    action = SyncAction.REPLACE if destination_path.exists() else SyncAction.COPY
    return SyncOperation(relative=rel_path, action=action)


async def _plan_sync_directories(
    source_root: Path,
    destination_root: Path,
    *,
    filters: SyncFilters,
    base: SyncBase,
    policy: SyncPolicy,
) -> AsyncIterator[SyncOperation]:
    source_prune_exclude = _prune_exclude_patterns(filters.source_exclude)
    destination_prune_exclude = _prune_exclude_patterns(filters.destination_exclude)

    async def process_directory_pair(
        source_dir: Path | None,
        dest_dir: Path | None,
        rel_dir: Path,
        *,
        source_pruned: bool,
        dest_pruned: bool,
    ) -> AsyncIterator[Emit[SyncOperation] | Spawn[SyncOperation]]:
        if source_dir is not None and not source_pruned:
            source_entries = await _scan_directory_single(source_dir, policy.semaphore)
        else:
            source_entries = _DirectoryEntries(files=set(), subdirs=set())

        if dest_dir is not None and not (policy.allow_delete and dest_pruned):
            dest_entries = await _scan_directory_single(dest_dir, policy.semaphore)
        else:
            dest_entries = _DirectoryEntries(files=set(), subdirs=set())

        for operation in _plan_file_propagation(
            rel_dir,
            source_entries.files,
            dest_entries.files,
            filters=filters,
            base=base,
        ):
            yield Emit(operation)

        if policy.allow_delete:
            for operation in _plan_file_cleanup(
                rel_dir,
                source_entries.files,
                dest_entries.files,
                filters.source_include,
                filters.source_exclude,
                base,
            ):
                yield Emit(operation)

        all_subdir_names = source_entries.subdirs | dest_entries.subdirs

        for subdir_name in all_subdir_names:
            subdir_rel_path = rel_dir / subdir_name

            in_source = subdir_name in source_entries.subdirs
            in_dest = subdir_name in dest_entries.subdirs

            source_subdir = (
                (source_dir / subdir_name) if (source_dir and in_source) else None
            )
            dest_subdir = (dest_dir / subdir_name) if (dest_dir and in_dest) else None

            sub_source_pruned = _should_prune_dir(
                subdir_rel_path, source_prune_exclude, base.source
            )
            sub_dest_pruned = _should_prune_dir(
                subdir_rel_path, destination_prune_exclude, base.destination
            )

            if sub_source_pruned and (not policy.allow_delete or sub_dest_pruned):
                continue

            yield Spawn(
                process_directory_pair(
                    source_subdir,
                    dest_subdir,
                    subdir_rel_path,
                    source_pruned=sub_source_pruned,
                    dest_pruned=sub_dest_pruned,
                )
            )

    initial_dest_root = destination_root if destination_root.exists() else None

    async for operation in stream_concurrently(
        process_directory_pair(
            source_root,
            initial_dest_root,
            Path("."),
            source_pruned=False,
            dest_pruned=False,
        )
    ):
        yield operation


async def _scan_directory_single(
    source_dir: Path,
    semaphore: asyncio.Semaphore,
) -> _DirectoryEntries:
    async with semaphore:
        source_subdirs, source_files = await asyncio.to_thread(
            _list_directory_entries, source_dir
        )

    return _DirectoryEntries(
        files={f.relative_to(source_dir) for f in source_files},
        subdirs={d.relative_to(source_dir) for d in source_subdirs},
    )


def _plan_file_propagation(
    rel_dir: Path,
    source_files: set[Path],
    dest_files: set[Path],
    *,
    filters: SyncFilters,
    base: SyncBase,
) -> Iterator[SyncOperation]:
    for file_name in source_files:
        file_rel_path = rel_dir / file_name

        if not _is_managed_with_base(
            file_rel_path, base.source, filters.source_include, filters.source_exclude
        ):
            continue

        if not _is_managed_with_base(
            file_rel_path,
            base.destination,
            filters.destination_include,
            filters.destination_exclude,
        ):
            yield SyncOperation(relative=file_rel_path, action=SyncAction.SKIP)
            continue

        action = SyncAction.REPLACE if file_name in dest_files else SyncAction.COPY
        yield SyncOperation(relative=file_rel_path, action=action)


def _plan_file_cleanup(
    rel_dir: Path,
    source_files: set[Path],
    dest_files: set[Path],
    source_include: tuple[str, ...],
    source_exclude: tuple[str, ...],
    base: SyncBase,
) -> Iterator[SyncOperation]:
    for file_name in dest_files - source_files:
        file_rel_path = rel_dir / file_name

        if _is_managed_with_base(
            file_rel_path, base.source, source_include, source_exclude
        ):
            yield SyncOperation(relative=file_rel_path, action=SyncAction.DELETE)


def _list_directory_entries(path: Path) -> tuple[list[Path], list[Path]]:
    subdirs: list[Path] = []
    files: list[Path] = []
    with os.scandir(path) as iterator:
        for entry in iterator:
            if entry.is_dir(follow_symlinks=False):
                subdirs.append(Path(entry.path))
            else:
                files.append(Path(entry.path))
    return subdirs, files


def _validate_sync_paths(source_path: Path, destination_path: Path) -> None:
    if _paths_overlap(source_path, destination_path):
        raise SyncError(
            "source and destination paths overlap: "
            f"{source_path} <-> {destination_path}"
        )
    if not source_path.exists():
        raise SyncError(f"source path not found: {source_path}")


def _paths_overlap(path_a: Path, path_b: Path) -> bool:
    return path_a == path_b or path_a in path_b.parents or path_b in path_a.parents


def _is_managed(
    rel_path: Path,
    include: tuple[str, ...],
    exclude: tuple[str, ...],
) -> bool:
    if exclude and any(_matches_pattern(rel_path, pattern) for pattern in exclude):
        return False
    return not include or any(
        _matches_pattern(rel_path, pattern) for pattern in include
    )


def _prune_exclude_patterns(exclude: tuple[str, ...]) -> tuple[str, ...]:
    if not exclude:
        return ()
    patterns: list[str] = []
    for pattern in exclude:
        if pattern.endswith("/**"):
            base_pattern = pattern[:-3]
            if base_pattern:
                patterns.append(base_pattern)
            continue
        if not _has_glob_magic(pattern):
            literal = pattern[:-1] if pattern.endswith("/") else pattern
            if literal:
                patterns.append(literal)
    return tuple(patterns)


def _should_prune_dir(
    rel_dir: Path,
    prune_exclude: tuple[str, ...],
    base: Path,
) -> bool:
    if not prune_exclude:
        return False
    match_path = base / rel_dir
    return any(_matches_pattern(match_path, pattern) for pattern in prune_exclude)


def _is_managed_with_base(
    rel_path: Path,
    base: Path,
    include: tuple[str, ...],
    exclude: tuple[str, ...],
) -> bool:
    return _is_managed(base / rel_path, include, exclude)


def _matches_pattern(rel_path: Path, pattern: str) -> bool:
    matcher = _compile_glob(pattern)
    return bool(matcher.match(rel_path.as_posix()))


def _has_glob_magic(pattern: str) -> bool:
    return any(char in pattern for char in "*?[")


@lru_cache(maxsize=256)
def _compile_glob(pattern: str) -> re.Pattern[str]:
    return re.compile(
        glob.translate(
            pattern,
            recursive=True,
            include_hidden=True,
            seps=("/",),
        )
    )



"""Public defaults that define contentctl's CLI contract."""

DEFAULT_CONFIG_FILENAME = "content-hub.yaml"
DEFAULT_CONFIG_PATH = DEFAULT_CONFIG_FILENAME
DEFAULT_INIT_DIR = "."
DEFAULT_WORKSPACE_SUBPATH = "."

__all__ = [
    "DEFAULT_CONFIG_FILENAME",
    "DEFAULT_CONFIG_PATH",
    "DEFAULT_INIT_DIR",
    "DEFAULT_WORKSPACE_SUBPATH",
]



from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping


def resolve_config(config: dict[str, Any], config_path: Path) -> ResolvedConfig:
    base_dir = config_path.parent
    defaults = config.get("defaults", {})
    default_include = _normalize_patterns(defaults.get("include", []))
    default_exclude = _normalize_patterns(defaults.get("exclude", []))

    origin = _resolve_workspace(
        base_dir=base_dir,
        default_include=default_include,
        default_exclude=default_exclude,
        name="",
        raw=config["origin"],
    )

    workspaces: dict[str, Workspace] = {}
    for name, raw in config["workspaces"].items():
        workspaces[name] = _resolve_workspace(
            base_dir=base_dir,
            default_include=default_include,
            default_exclude=default_exclude,
            name=name,
            raw=raw,
        )

    return ResolvedConfig(origin=origin, workspaces=workspaces)


@dataclass(frozen=True)
class Workspace:
    name: str
    path: Path
    include: tuple[str, ...]
    exclude: tuple[str, ...]


@dataclass(frozen=True)
class ResolvedConfig:
    origin: Workspace
    workspaces: dict[str, Workspace]


def _resolve_workspace(
    base_dir: Path,
    default_include: list[str],
    default_exclude: list[str],
    name: str,
    raw: Any,
) -> Workspace:
    raw_value = cast("dict[str, Any] | str", raw)
    match raw_value:
        case str():
            resolved_path = _resolve_path(base_dir, raw_value)
            include = tuple(default_include)
            exclude = tuple(default_exclude)
        case dict():
            raw_dict = cast("Mapping[str, Any]", raw_value)
            path_value = cast("str", raw_dict["path"])
            include_raw = cast("list[str]", raw_dict.get("include", []))
            exclude_raw = cast("list[str]", raw_dict.get("exclude", []))
            resolved_path = _resolve_path(base_dir, path_value)
            include = tuple(_normalize_patterns(default_include, include_raw))
            exclude = tuple(_normalize_patterns(default_exclude, exclude_raw))
    return Workspace(name=name, path=resolved_path, include=include, exclude=exclude)


def _resolve_path(base_dir: Path, value: str) -> Path:
    path = Path(value)
    if not path.is_absolute():
        path = base_dir / path
    return path.resolve()


def _normalize_patterns(*pattern_groups: Iterable[str]) -> list[str]:
    patterns: list[str] = []
    for group in pattern_groups:
        for pattern in group:
            cleaned = pattern.strip().replace("\\", "/")
            if cleaned.startswith("./"):
                cleaned = cleaned[2:]
            if cleaned:
                patterns.append(cleaned)
    return patterns
